# Eeterm -- Terminal for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eeterm
#
# This file is part of Eeterm.
#
# Eeterm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Eeterm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Core Eeterm stuff"""

import logging
import threading

from eessh.api import api
from eessh.config import config as eessh_config
from gi.repository import Gdk, Gio, GLib, Gtk

from eeterm.config import config as eeterm_config
from eeterm.const import APPLICATION_ID, UI_FILEPATH
from eeterm.term import Terminal, get_colors, get_palette, hex2rgba, rgb2hex

log = logging.getLogger(__name__)


TAB_POSITIONS = {
    "left": Gtk.PositionType.LEFT,
    "right": Gtk.PositionType.RIGHT,
    "top": Gtk.PositionType.TOP,
    "bottom": Gtk.PositionType.BOTTOM,
}


CONFIGS = {
    "eessh": eessh_config,
    "eeterm": eeterm_config,
}


class Application(Gtk.Application):
    """Eeterm application"""

    def __init__(self):
        # Calling inherited
        super().__init__(application_id=APPLICATION_ID, flags=Gio.ApplicationFlags.NON_UNIQUE)

        self.builder = None

        self.pref_entries = []
        self.terminals = []

        self.connect("activate", self.on_activate)
        self.connect("shutdown", self.on_shutdown)

        self.refresh_handler_id = None

    def add_terminal(self, copy=False, data=None, position=-1):
        """Instantiate and run a terminal in a new tab (=page)"""
        if data:
            # New terminal with client
            terminal = Terminal(self.builder, data=data)
        else:
            # Get current terminal
            terminal = self.get_current_terminal()

            if terminal and (terminal.is_local or copy):
                # Copy or new terminal
                terminal = terminal.copy()
            else:
                # New terminal only
                terminal = Terminal(self.builder)

        # Scrollbar
        scrollbar = Gtk.VScrollbar()
        scrollbar.set_adjustment(terminal.get_vadjustment())

        hbox = Gtk.HBox()
        hbox.pack_start(terminal, True, True, 0)
        hbox.pack_start(scrollbar, False, False, 0)

        # Remove the notebook page
        terminal.connect("child-exited", self.on_terminal_child_exited)

        # Track terminal stage
        terminal.connect("child-ready", self.on_terminal_child_ready)

        # Track terminal changes
        terminal.connect("contents-changed", self.on_terminal_contents_changed)

        self.insert_page(hbox, terminal.label, position)

        # Give the focus to the terminal
        terminal.set_property("has-focus", True)

        # Go!
        terminal.start()

        # Register terminal
        self.terminals.append(terminal)

    def get_current_terminal(self):
        """Get current terminal"""

        # Get current page
        notebook = self.builder.get_object("main-notebook")
        page_num = notebook.get_current_page()
        if page_num < 1:
            # Not a terminal
            return None

        # Get terminal
        hbox = notebook.get_nth_page(page_num)
        return hbox.get_children()[0]

    def hide_refresh(self, error=False):
        """Hide refresh popup window"""

        if self.refresh_handler_id:
            GLib.source_remove(self.refresh_handler_id)
            self.refresh_handler_id = None

        refresh_window = self.builder.get_object("refresh-window")
        refresh_window.hide()

        if error:
            self.show_alert(
                "Error",
                "<b>Error:</b> Failed to get data from remote server.\n\n"
                "Please check API parameters and/or network connectivity.",
            )

    def insert_page(self, widget, label, position):
        """Add a new tab (=page)"""
        notebook = self.builder.get_object("main-notebook")

        # Pack label and close button
        box = Gtk.Box(Gtk.Orientation.HORIZONTAL)
        box.pack_start(label, False, True, 10)  # pylint: disable=no-member

        if position == -1 or position > 0:
            # Create close button
            button = Gtk.Button.new_from_icon_name(
                "gtk-close", Gtk.IconSize.MENU  # pylint: disable=no-member
            )
            button.set_relief(Gtk.ReliefStyle.NONE)
            button.connect("clicked", self.on_close_button_activate, widget)

            box.pack_start(button, False, False, 0)

        # Insert the page
        page_num = notebook.insert_page(widget, box, position=position)

        notebook.set_tab_reorderable(widget, True)
        notebook.set_tab_detachable(widget, True)

        box.show_all()
        widget.show_all()

        notebook.set_current_page(page_num)

        return page_num

    def on_about_activate(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on about button click: run the about dialog"""
        dialog = self.builder.get_object("about-dialog")
        dialog.run()
        dialog.hide()

    # pylint: disable=unused-argument
    @staticmethod
    def _get_filter_set_visible_func(entry, treeview):
        def set_visible(model, iter_, data=None, entry=entry, treeview=treeview):
            query = entry.get_text().lower()
            if not query:
                return True

            for column in range(1, treeview.get_n_columns()):
                value = model.get_value(iter_, column)
                if isinstance(value, str) and query in value.lower():
                    return True

            return False

        return set_visible

    # pylint: disable=unused-argument
    def on_activate(self, app):
        """Main callback on application starting"""
        self.builder = Gtk.Builder()
        self.builder.add_from_file(UI_FILEPATH)
        self.builder.connect_signals(self)
        self.builder.set_application(self)  # pylint: disable=no-member

        window = self.builder.get_object("application-window")
        window.set_title(eeterm_config.appname)
        window.connect("destroy", self.on_quit_activate)
        window.show()

        # Connect toolbar menu
        menu = self.builder.get_object("connect-menu")
        button = self.builder.get_object("connect-toolbutton")
        button.set_menu(menu)

        # keyboard shortcuts
        accel = Gtk.AccelGroup()
        accel.connect(
            Gdk.keyval_from_name("M"),
            Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK,
            0,
            self.on_toggle_menu_activate,
        )
        window.add_accel_group(accel)

        self.add_window(window)

        notebook = self.builder.get_object("main-notebook")
        notebook.set_tab_pos(TAB_POSITIONS[eeterm_config.get("global", "tab_position")])

        def add_treeview_columns(treeview, columns):
            for index, (type_, title, alignment, visible) in enumerate(columns):
                if type_ in (int, str):
                    renderer = Gtk.CellRendererText()
                    if alignment == "right":
                        renderer.set_alignment(1.0, 0.5)
                    column = Gtk.TreeViewColumn(title, renderer, text=index)
                elif type_ == bool:
                    renderer = Gtk.CellRendererToggle()
                    column = Gtk.TreeViewColumn(title, renderer, active=index)

                column.set_sort_column_id(index)
                column.set_visible(visible)

                treeview.append_column(column)

        # Customers treeview
        columns = [
            (int, "Id", None, False),
            (str, "EeId", None, True),
            (str, "Label", None, True),
        ]

        treeview = self.builder.get_object("customers-treeview")
        add_treeview_columns(treeview, columns)

        model = Gtk.TreeStore(*[column[0] for column in columns])
        filter_ = model.filter_new()

        treeview.set_model(filter_)
        treeview.set_search_column(1)

        entry = self.builder.get_object("customers-search_entry")
        entry.set_property("has-focus", True)

        filter_.set_visible_func(self._get_filter_set_visible_func(entry, treeview))

        # Servers treeview
        columns = [
            (int, "Id", None, False),
            (str, "Hostname", None, True),
            (str, "Hoster", None, False),
            (str, "Operating System", None, True),
            (str, "Usage", None, True),
            (str, "Description", None, True),
            (object, "data", None, False),
        ]

        treeview = self.builder.get_object("servers-treeview")
        add_treeview_columns(treeview, columns)

        model = Gtk.TreeStore(*[column[0] for column in columns])
        filter_ = model.filter_new()

        treeview.set_model(filter_)
        treeview.set_search_column(1)

        entry = self.builder.get_object("servers-search_entry")

        filter_.set_visible_func(self._get_filter_set_visible_func(entry, treeview))

        # PFWDs treeview
        columns = [
            (int, "Index", None, False),
            (str, "Server host", None, True),
            (int, "Server port", "right", True),
            (str, "Remote host", None, True),
            (int, "Remote port", "right", True),
        ]

        treeview = self.builder.get_object("pfwds-treeview")
        add_treeview_columns(treeview, columns)

        model = Gtk.TreeStore(*[column[0] for column in columns])
        treeview.set_model(model)
        treeview.set_search_column(2)

        # Connect preference entries «changed» signals
        def connect_entry(name, config):
            entry = self.builder.get_object(name)
            assert entry is not None, f"Invalid preference entry name {name}"

            section, option, type_ = name.split("-")
            entry.config = config
            entry.section = section
            entry.option = option

            if type_ == "check_button":
                entry.get_value = entry.get_active
                entry.set_value = entry.set_active
                signal_name = "toggled"

            elif type_ == "color_button":

                def get_value(entry=entry):
                    rgba = entry.get_rgba()
                    return rgb2hex(rgba.red, rgba.green, rgba.blue)

                def set_value(value, entry=entry):
                    if not value:
                        return
                    rgba = Gdk.RGBA(*hex2rgba(value))
                    entry.set_rgba(rgba)

                entry.get_value = get_value
                entry.set_value = set_value
                signal_name = "color-set"

            elif type_ == "combobox_text":
                entry.get_value = entry.get_active_id
                entry.set_value = entry.set_active_id
                signal_name = "changed"

            elif type_ == "entry":
                entry.get_value = entry.get_text
                entry.set_value = entry.set_text
                signal_name = "changed"

            elif type_ == "font_button":
                entry.get_value = entry.get_font_name
                entry.set_value = entry.set_font_name
                signal_name = "font-set"

            elif type_ == "spin_button":
                signal_name = "value-changed"

            elif type_ == "switch":
                entry.get_value = entry.get_active
                entry.set_value = entry.set_active
                signal_name = "state-set"

            entry.handler_id = entry.connect(signal_name, self.on_preference_changed)

            self.pref_entries.append(entry)

        # API
        for name in (
            "legacy-switch",
            "url-entry",
            "username-entry",
            "password-entry",
            "debug-switch",
        ):
            connect_entry(f"api-{name}", "eessh")

        # Connection
        for name in ("agent_forwarding-switch", "compress-switch", "x11_forwarding-switch"):
            connect_entry(f"connection-{name}", "eessh")

        # Firsthop
        for name in (
            "enabled-switch",
            "hostname-entry",
            "username-entry",
            "password-entry",
            "port-spin_button",
        ):
            connect_entry(f"firsthop-{name}", "eessh")

        # Global
        for name in ("tab_position-combobox_text",):
            connect_entry(f"global-{name}", "eeterm")

        # Terminal
        for name in (
            "use_system_font-switch",
            "font-font_button",
            "allow_bold-switch",
            "audible_bell-switch",
            "mouse_autohide-switch",
            "word_char_exceptions-entry",
            "cursor_blink_mode-combobox_text",
            "prefered_shell-entry",
            "backspace_binding-combobox_text",
            "delete_binding-combobox_text",
        ):
            connect_entry(f"terminal-{name}", "eeterm")

        # Scrolling
        for name in (
            "scrollback_lines-spin_button",
            "scroll_on_output-switch",
            "scroll_on_keystroke-switch",
            "scroll_step_increment-spin_button",
            "scroll_page_increment-spin_button",
        ):
            connect_entry(f"terminal-{name}", "eeterm")

        # Colors
        for name in (
            "color_scheme-combobox_text",
            "foreground-color_button",
            "background-color_button",
            "color_bold-color_button",
            "palette_scheme-combobox_text",
        ):
            connect_entry(f"terminal-{name}", "eeterm")

        # Custom palette
        for index in range(0, 16):
            connect_entry(f"terminal-palette_{index}-color_button", "eeterm")

        # Get customers list
        GLib.timeout_add(100, self.on_refresh_clicked)

    def on_become_root_clicked(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on become root button clicked: Telling current terminal to become root"""

        # Get terminal
        terminal = self.get_current_terminal()
        if terminal is None:
            return

        terminal.become_root()

    def on_broadcast_toggled(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on broadcast button clicked: Toggle broadcast mode on current terminal"""

        # Get terminal
        terminal = self.get_current_terminal()
        if terminal is None:
            return

        terminal.toggle_broadcast()

        if terminal.broadcast == "master":
            # Wait for fifo files to be setup then toggle other terminals
            def propagate(terminal):
                for other in self.terminals:
                    if other == terminal:
                        continue

                    other.toggle_broadcast()

            GLib.timeout_add(100, propagate, terminal)

    def on_close_page_activate(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on closing tab (=page)"""
        # Get current page
        notebook = self.builder.get_object("main-notebook")
        page_num = notebook.get_current_page()
        if page_num < 1:
            # Not a terminal
            return

        hbox = notebook.get_nth_page(page_num)

        self.on_close_button_activate(None, hbox)

    def on_close_button_activate(self, button, hbox):  # pylint: disable=unused-argument
        """Callback on tab close button click: exit terminal and remove the page"""
        # Close the terminal
        terminal = hbox.get_children()[0]
        if terminal.on_child_quit():
            # The page will be removed by the later callback
            return

        # Failed kill
        notebook = self.builder.get_object("main-notebook")
        page_num = notebook.page_num(hbox)
        notebook.remove_page(page_num)

    def on_connect_clicked(self, button):  # pylint: disable=unused-argument
        """Callback on connect button click: Connect to selected servers"""

        treeview = self.builder.get_object("servers-treeview")
        treeselection = treeview.get_selection()
        _, paths = treeselection.get_selected_rows()

        for path in paths:
            self.on_servers_treeview_row_activated(treeview, path, 0)

    def on_connect_with_agent_forwarding_activate(self, button):  # pylint: disable=unused-argument
        """
        Callback on connect with agent menu item click:
        Connect to selected servers with agent forwarding
        """

        options = {"agent_forwarding": True}

        treeview = self.builder.get_object("servers-treeview")
        treeselection = treeview.get_selection()
        _, paths = treeselection.get_selected_rows()

        for path in paths:
            self.on_servers_treeview_row_activated(treeview, path, 0, **options)

    def on_connect_with_options_activate(self, button):  # pylint: disable=unused-argument
        """
        Callback on connect with options menu item click:
        Connect to selected servers with options dialog
        """

        # Set default values
        for option in ("agent_forwarding", "compress", "x11_forwarding"):
            value = eessh_config.get("connection", option)
            self.builder.get_object(f"connect-{option}-switch").set_active(value)

        self.builder.get_object("connect-firsthop_enabled-switch").set_active(
            eessh_config.get("firsthop", "enabled")
        )

        # SSH user
        treeview = self.builder.get_object("servers-treeview")
        treeselection = treeview.get_selection()
        model, paths = treeselection.get_selected_rows()

        usernames = set()
        for index, path in enumerate(paths):
            iter_ = model.get_iter(path)
            data = model.get_value(iter_, 6)

            if index == 0:
                usernames.update(data["usernames"])
            else:
                usernames.intersection_update(data["usernames"])

        combo = self.builder.get_object("connect-username-combo_box_text")
        model = combo.get_model()
        model.clear()

        combo.append(None, "Default user")
        combo.set_active(0)

        for username in sorted(usernames):
            combo.append(username, username)

        # Run the dialog
        dialog = self.builder.get_object("connect-dialog")
        dialog.show()
        response = dialog.run()
        dialog.hide()

        if response != Gtk.ResponseType.OK:
            return

        # Get options
        options = {}
        for option in ("agent_forwarding", "compress", "x11_forwarding"):
            value = self.builder.get_object(f"connect-{option}-switch").get_active()
            options[option] = value

        options["no_firsthop"] = not self.builder.get_object(
            "connect-firsthop_enabled-switch"
        ).get_active()
        options["username"] = self.builder.get_object(
            "connect-username-combo_box_text"
        ).get_active_id()

        treeview = self.builder.get_object("servers-treeview")
        treeselection = treeview.get_selection()
        _, paths = treeselection.get_selected_rows()

        for path in paths:
            self.on_servers_treeview_row_activated(treeview, path, 0, **options)

    def on_connect_without_firsthop_activate(self, button):  # pylint: disable=unused-argument
        """
        Callback on connect without first hop menu item click:
        Connect to selected servers without first hop
        """

        options = {"no_firsthop": True}

        treeview = self.builder.get_object("servers-treeview")
        treeselection = treeview.get_selection()
        _, paths = treeselection.get_selected_rows()

        for path in paths:
            self.on_servers_treeview_row_activated(treeview, path, 0, **options)

    def on_copy_clicked(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on copy terminal button click: Copy current terminal"""

        self.add_terminal(copy=True)

    def on_customers_search_entry_search_changed(self, *args):  # pylint: disable=unused-argument
        """Callback on search customer pattern change: Apply filter to customers list"""

        treeview = self.builder.get_object("customers-treeview")
        filter_ = treeview.get_model()
        filter_.refilter()

    def on_customers_treeview_row_activated(
        self, treeview, path, column
    ):  # pylint: disable=unused-argument
        """Callback on customers list item clicked"""
        model = treeview.get_model()
        iter_ = model.get_iter(path)
        customer_uid = model.get_value(iter_, 1)

        def refresh_model(data):
            if data is None:
                self.hide_refresh(error=True)
                return

            treeview = self.builder.get_object("servers-treeview")

            filter_ = treeview.get_model()
            model = filter_.get_model()
            model.clear()

            for item in data:
                log.debug(item)
                model.append(
                    None,
                    [
                        int(item["id"]),
                        item["hostname"],
                        item["hoster"],
                        item["os_name"],
                        item["usage"],
                        item["description"],
                        {
                            "usernames": item["usernames"],
                        },
                    ],
                )

            self.hide_refresh()

        def call_api(customer_uid):
            data = api.get_servers(customer_uid)
            GLib.idle_add(refresh_model, data)

        self.show_refresh()

        thread = threading.Thread(target=call_api, args=(customer_uid,))
        thread.daemon = True
        thread.start()

    def on_move_left_activate(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on moving a tab (=page) to left menu item click"""
        notebook = self.builder.get_object("main-notebook")
        page_num = notebook.get_current_page()
        if page_num < 2:
            return

        page = notebook.get_nth_page(page_num)
        notebook.reorder_child(page, page_num - 1)

    def on_move_right_activate(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on moving a tab (=page) to right"""
        notebook = self.builder.get_object("main-notebook")
        page_num = notebook.get_current_page()
        if page_num < 1:
            return

        page = notebook.get_nth_page(page_num)
        notebook.reorder_child(page, page_num + 1)

    def on_next_page_activate(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback next tab menu item click"""
        notebook = self.builder.get_object("main-notebook")
        notebook.next_page()

    def on_new_terminal_activate(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on new terminal menu item click: add a new terminal"""

        self.add_terminal()

    def on_new_terminal_clicked(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on new terminal button click: add a new terminal"""

        self.add_terminal()

    def on_pfwd_delete_clicked(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on port forwarding delete button click: Close a port forwarding tunnel"""

        terminal = self.get_current_terminal()
        if terminal is None or terminal.is_local:
            return

        treeview = self.builder.get_object("pfwds-treeview")
        selection = treeview.get_selection()
        model, iter_ = selection.get_selected()
        if iter_ is None:
            return

        index = model.get_value(iter_, 0)

        dialog = self.builder.get_object("pfwd-manage-dialog")

        def callback(result, dialog=dialog):
            if result:
                dialog.callback()
                return

            self.show_alert(
                "Error",
                "<b>Error:</b> Failed to close tunnel.\n\n"
                "Please check console and/or logs for details.",
            )

        if dialog.mode == "pfwd":
            terminal.close_pfwd(index, callback=callback)
        else:
            terminal.close_rfwd(index, callback=callback)

    def on_pfwd_create_clicked(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on port forwarding add button click: Add a new port forwarding"""

        self.pfwd_create()

    def on_pfwd_manage_clicked(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on port forwarding management button click: open port forwarding dialog"""

        self.pfwd_manage("pfwd")

    def on_pfwds_treeview_selection_changed(self, treeselection):
        """Change «connect» button sensitivity"""

        button = self.builder.get_object("pfwd_delete-button")
        button.set_sensitive(treeselection.count_selected_rows() > 0)

    def on_preference_changed(self, entry, *args, init=False):  # pylint: disable=unused-argument
        """Preference entries callback for «changed»  signal"""

        if not init:
            # Save configuration
            CONFIGS[entry.config].set(
                entry.section,
                entry.option,
                entry.get_value(),
            )
            CONFIGS[entry.config].save()

        # Extra actions
        if entry.section == "api":
            if not init:
                api.initialize(eessh_config)

        elif entry.section == "terminal":
            if entry.option == "color_scheme":
                # Set foreground, background and color_bold color button values from selected scheme
                # Enable buttons only if scheme is «custom»
                scheme = entry.get_value()
                colors = get_colors(scheme)

                for index, option in enumerate(("foreground", "background", "color_bold")):
                    color_button = self.builder.get_object(f"terminal-{option}-color_button")

                    if colors[index]:
                        color_button.handler_block(color_button.handler_id)
                        color_button.set_rgba(colors[index])
                        color_button.handler_unblock(color_button.handler_id)

                    color_button.set_sensitive(scheme == "custom")

            elif entry.option == "palette_scheme":
                # Set palette color buttons values from selected palette
                # Enable buttons only if scheme is «custom»
                scheme = entry.get_active_id()
                palette = get_palette(scheme)

                for index in range(0, 16):
                    color_button = self.builder.get_object(f"terminal-palette_{index}-color_button")

                    if palette:
                        color_button.handler_block(color_button.handler_id)
                        color_button.set_rgba(palette[index])
                        color_button.handler_unblock(color_button.handler_id)

                    color_button.set_sensitive(scheme == "custom")

            elif entry.option == "use_system_font":
                # Enable font button only if not set
                font_button = self.builder.get_object("terminal-font-font_button")
                font_button.set_sensitive(not entry.get_value())

            if not init:
                # Reload terminals configuration
                for terminal in self.terminals:
                    terminal.load_config()

        elif entry.section == "firsthop":
            if entry.option == "enabled":
                for option in ("hostname", "username", "password"):
                    entry_ = self.builder.get_object(f"firsthop-{option}-entry")
                    entry_.set_sensitive(entry.get_value())

                spin_button = self.builder.get_object("firsthop-port-spin_button")
                spin_button.set_sensitive(entry.get_value())

        elif entry.section == "global":
            if entry.option == "tab_position" and not init:
                # Change main notebook tab position
                notebook = self.builder.get_object("main-notebook")
                notebook.set_tab_pos(TAB_POSITIONS[entry.get_value()])

    def on_preferences_activate(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Run preference dialog and save configurations"""

        # Set values to entries
        for entry in self.pref_entries:
            value = CONFIGS[entry.config].get(entry.section, entry.option)
            entry.handler_block(entry.handler_id)
            entry.set_value(value if value is not None else "")
            entry.handler_unblock(entry.handler_id)

            self.on_preference_changed(entry, init=True)

        # Show the dialog
        dialog = self.builder.get_object("preferences-dialog")
        dialog.run()
        dialog.hide()

        # Save configurations
        for config in CONFIGS.values():
            config.save()

    def on_prev_page_activate(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on previous tab menu item click"""
        notebook = self.builder.get_object("main-notebook")
        notebook.prev_page()

    def on_quit_activate(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Exit application by calling Gtk.Application.quit()"""

        # Close all termrinals
        for terminal in self.terminals:
            terminal.on_child_quit()

        self.quit()

    def on_refresh_clicked(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on refresh button click"""
        if not api.ready:
            self.show_alert("Error", "<b>Error:</b> Please configure API parameters.")
            return

        def refresh_model(data):
            if data is None:
                self.hide_refresh(error=True)
                return

            treeview = self.builder.get_object("customers-treeview")

            filter_ = treeview.get_model()
            model = filter_.get_model()
            model.clear()

            for item in data:
                model.append(None, [1, item["customer_uid"], item["name"]])

            treeview = self.builder.get_object("servers-treeview")

            filter_ = treeview.get_model()
            model = filter_.get_model()
            model.clear()

            self.hide_refresh()

        def call_api():
            data = api.get_customers()
            GLib.idle_add(refresh_model, data)

        self.show_refresh()

        thread = threading.Thread(target=call_api)
        thread.daemon = True
        thread.start()

    def on_rfwd_manage_clicked(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Reverse port forwarding from toolbar"""

        self.pfwd_manage("rfwd")

    def on_shutdown(self, app):  # pylint: disable=unused-argument
        """Called by Gtk.Application.quit()"""

    def on_servers_search_entry_search_changed(self, *args):  # pylint: disable=unused-argument
        """Apply filter to servers list"""

        treeview = self.builder.get_object("servers-treeview")
        filter_ = treeview.get_model()
        filter_.refilter()

    def on_servers_treeview_button_press_event(self, treeview, event):
        """Show connect menu on right click"""

        if event.button != 3:
            return None

        result = treeview.get_path_at_pos(event.x, event.y)
        if result is None:
            return None

        # Show context menu
        menu = self.builder.get_object("connect-menu")
        menu.popup(None, None, None, None, event.button, activate_time=event.time)

        # Prevent from altering selection
        treeselection = treeview.get_selection()
        _, paths = treeselection.get_selected_rows()

        return result[0] in paths

    def on_servers_treeview_row_activated(
        self, treeview, path, column, **options
    ):  # pylint: disable=unused-argument
        """Callback on serves list item click"""
        model = treeview.get_model()
        iter_ = model.get_iter(path)
        server_id = model.get_value(iter_, 0)

        def add_terminal(data):
            if data is None:
                self.hide_refresh(error=True)
                return

            data.update(options)

            self.hide_refresh()

            self.add_terminal(data=data)

        def call_api(server_id):
            data = api.get_data(server_id=server_id)
            GLib.idle_add(add_terminal, data)

        self.show_refresh()

        thread = threading.Thread(target=call_api, args=(server_id,))
        thread.daemon = True
        thread.start()

    def on_servers_treeview_selection_changed(
        self, treeselection
    ):  # pylint: disable=unused-argument
        """Change «connect» button sensitivity"""

        button = self.builder.get_object("connect-toolbutton")
        button.set_sensitive(treeselection.count_selected_rows() > 0)

    def on_main_notebook_switch_page(
        self, notebook, page, page_num
    ):  # pylint: disable=unused-argument
        """Called when user switch notebook page"""

        if page_num > 0:
            # Terminal
            terminal = page.get_children()[0]

            # Unhighlight the label
            text = terminal.label.get_text()
            terminal.label.set_text(text)

            # Update buttons sensitivity
            for name in ("copy", "pfwd_manage", "rfwd_manage"):
                button = self.builder.get_object(f"{name}-toolbutton")
                button.set_sensitive(not terminal.is_local)

            for name in ("become-root", "transfer"):
                button = self.builder.get_object(f"{name}-toolbutton")
                button.set_sensitive(not terminal.is_local and terminal.mode == "shell")

            # Special case for broadcast button
            button = self.builder.get_object("broadcast-toolbutton")
            button.handler_block_by_func(self.on_broadcast_toggled)
            button.set_active(terminal.broadcast != "off")
            button.handler_unblock_by_func(self.on_broadcast_toggled)
            button.set_sensitive(not terminal.is_local and terminal.mode == "shell")

            # Connect button
            button = self.builder.get_object("connect-toolbutton")
            button.set_sensitive(False)

        else:
            # Not a terminal
            for name in (
                "copy",
                "become-root",
                "broadcast",
                "pfwd_manage",
                "transfer",
                "rfwd_manage",
            ):
                button = self.builder.get_object(f"{name}-toolbutton")
                button.set_sensitive(False)

            # Connect button
            treeview = self.builder.get_object("servers-treeview")
            selection = treeview.get_selection()
            selection.emit("changed")

    def on_terminal_child_exited(
        self, terminal, *args, **kwargs
    ):  # pylint: disable=unused-argument
        """Remove the page"""

        # Remove the page
        page = terminal.get_parent()

        notebook = self.builder.get_object("main-notebook")
        page_num = notebook.page_num(page)

        log.debug("child-exited signal was emitted by terminal on page %s", page_num)

        notebook.remove_page(page_num)
        self.terminals.remove(terminal)

    def on_terminal_child_ready(self, terminal):  # pylint: disable=unused-argument
        """Callback when a child terminal is ready: re-enable toolbar buttons"""

        for name in ("become-root", "broadcast", "transfer"):
            button = self.builder.get_object(f"{name}-toolbutton")
            button.set_sensitive(not terminal.is_local)

    def on_terminal_contents_changed(self, terminal):  # pylint: disable=unused-argument
        """Callback on terminal contents change"""
        notebook = self.builder.get_object("main-notebook")

        page = terminal.get_parent()
        page_num = notebook.page_num(page)
        if page_num != notebook.get_current_page():
            # Highlight the label if not current terminal
            text = terminal.label.get_text()
            terminal.label.set_markup(f'<span foreground="#CC3333">{text}</span>')

    def on_toggle_menu_activate(
        self, accel, window, signal_id, modifiers
    ):  # pylint: disable=unused-argument
        """Callback on toggle menu bar visibility shortcut: Toggle menu bar visibility"""

        menubar = self.builder.get_object("main-menubar")
        if menubar.is_visible():
            menubar.hide()
        else:
            menubar.show()

    def on_transfer_clicked(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Callback on toolbar transfer button click"""

        # Not implemented
        return
        # pylint: disable=unreachable
        terminal = self.get_current_terminal()
        if terminal is None:
            return

        if terminal.is_local:
            # TODO: popup
            return

        filechooser_dialog = self.builder.get_object("filechooser-dialog")
        response = filechooser_dialog.run()
        filechooser_dialog.hide()

        if response == Gtk.ResponseType.OK:
            if not terminal.send_files(filechooser_dialog.get_filenames()):
                # TODO: popup
                return

    def on_zoom_in_activate(self, *args):  # pylint: disable=unused-argument
        """Callback on zoom in terminal menu item click"""

        # Get terminal
        terminal = self.get_current_terminal()
        if terminal is None:
            return

        terminal.zoom_in()

    def on_zoom_out_activate(self, *args):  # pylint: disable=unused-argument
        """Callback on zoom out terminal menu item click"""

        # Get terminal
        terminal = self.get_current_terminal()
        if terminal is None:
            return

        terminal.zoom_out()

    def on_zoom_reset_activate(self, *args):  # pylint: disable=unused-argument
        """Callback on reset zoom menu item click"""

        # Get terminal
        terminal = self.get_current_terminal()
        if terminal is None:
            return

        terminal.zoom_reset()

    def pfwd_create(self):
        """
        Open a port forwarding tunnel
        Display dialog and handle what happens next
        """

        terminal = self.get_current_terminal()
        if terminal is None or terminal.is_local:
            return

        # Set default values
        self.builder.get_object("pfwd-server_host-entry").set_text("127.0.0.1")
        self.builder.get_object("pfwd-server_port-spin_button").set_value(8080)
        self.builder.get_object("pfwd-remote_host-entry").set_text("127.0.0.1")
        self.builder.get_object("pfwd-remote_port-spin_button").set_value(80)

        # Run the dialog
        dialog = self.builder.get_object("pfwd-create-dialog")

        # Set title
        if dialog.mode == "pfwd":
            # PFWD
            dialog.set_title("Open port forwarding tunnel with remote")
        else:
            # RFWD
            dialog.set_title("Open reverse port forwarding tunnel with remote")

        dialog.show()
        response = dialog.run()
        dialog.hide()

        if response != Gtk.ResponseType.OK:
            return

        # Get settings
        server_host = self.builder.get_object("pfwd-server_host-entry").get_text()
        server_port = int(self.builder.get_object("pfwd-server_port-spin_button").get_value())
        remote_host = self.builder.get_object("pfwd-remote_host-entry").get_text()
        remote_port = int(self.builder.get_object("pfwd-remote_port-spin_button").get_value())

        def callback(result, dialog=dialog):
            if result:
                dialog.callback()
                return

            self.show_alert(
                "Error",
                "<b>Error:</b> Failed to open tunnel.\n\n"
                "Please check console and/or logs for details.",
            )

        # Open tunnel
        if dialog.mode == "pfwd":
            # PFWD
            terminal.open_pfwd(
                server_host, server_port, remote_host, remote_port, callback=callback
            )
        else:
            # RFWD
            terminal.open_rfwd(
                server_host, server_port, remote_host, remote_port, callback=callback
            )

    def pfwd_manage(self, mode):
        """
        Manage forwarding tunnels
        Open port forwarning dialog
        """

        terminal = self.get_current_terminal()
        if terminal is None or terminal.is_local:
            return

        def refresh_pfwds_treeview(mode=mode, terminal=terminal):
            # Refresh opened tunnels list
            treeview = self.builder.get_object("pfwds-treeview")
            model = treeview.get_model()
            model.clear()

            if mode == "pfwd":
                pfwds = terminal.pfwds
            else:
                pfwds = terminal.rfwds

            for index, server_host, server_port, remote_host, remote_port in pfwds:
                model.append(None, [index, server_host, server_port, remote_host, remote_port])

        refresh_pfwds_treeview()

        # Set mode to create dialog
        dialog = self.builder.get_object("pfwd-create-dialog")
        dialog.callback = refresh_pfwds_treeview
        dialog.mode = mode

        # Run manage dialog
        dialog = self.builder.get_object("pfwd-manage-dialog")
        dialog.callback = refresh_pfwds_treeview
        dialog.mode = mode

        if mode == "pfwd":
            dialog.set_title("Port forwarding tunnels management")
        else:
            dialog.set_title("Reverse port forwarding tunnels management")

        dialog.show()
        dialog.run()
        dialog.hide()

    def show_alert(self, title, message):
        """Show alert dialog"""

        dialog = self.builder.get_object("alert-dialog")
        dialog.set_title(title)
        dialog.set_markup(message)
        dialog.run()
        dialog.hide()

    def show_refresh(self):
        """Show refresh popup window"""

        refresh_window = self.builder.get_object("refresh-window")
        refresh_window.present()

        def do_pulse():
            progressbar = self.builder.get_object("refresh-progressbar")
            progressbar.pulse()
            return True

        self.refresh_handler_id = GLib.timeout_add(75, do_pulse)
