# Eeterm -- Terminal for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eeterm
#
# This file is part of Eeterm.
#
# Eeterm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Eeterm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Terminal stuff"""

import logging
import os
import pty
import signal
import threading

from eessh.client import EeSSHClient
from eessh.config import config as eessh_config
from gi.repository import Gdk, GLib, GObject, Gtk, Pango, Vte

from eeterm.config import config as eeterm_config

log = logging.getLogger(__name__)

CONFIGS = {
    "eessh": eessh_config,
    "eeterm": eeterm_config,
}

CURSOR_BLINK_MODES = {
    "on": Vte.CursorBlinkMode.ON,
    "off": Vte.CursorBlinkMode.OFF,
    "system": Vte.CursorBlinkMode.SYSTEM,
}

ERASE_BINDINGS = {
    "auto": Vte.EraseBinding.AUTO,
    "delete": Vte.EraseBinding.ASCII_DELETE,
    "backspace": Vte.EraseBinding.ASCII_BACKSPACE,
}

COLOR_SCHEMES = {
    "black-on-light-yellow": ("#000000", "#FFFFDB", "#000000"),
    "black-on-white": ("#000000", "#FFFFFF", "#000000"),
    "gray-on-black": ("#A8A8A8", "#000000", "#A8A8A8"),
    "green-on-black": ("#00FF00", "#000000", "#00FF00"),
    "white-on-black": ("#FFFFFF", "#000000", "#FFFFFF"),
}

PALETTE_SCHEMES = {
    "xterm": (
        "#000000",
        "#CC0000",
        "#00CC00",
        "#CCCC00",
        "#1E8EFF",
        "#CC00CC",
        "#00CCCC",
        "#E5E5E5",
        "#4C4C4C",
        "#FF0000",
        "#00FF00",
        "#FFFF00",
        "#4482B2",
        "#FF00FF",
        "#00FFFF",
        "#FFFFFF",
    ),
    "rxvt": (
        "#000000",
        "#CC0000",
        "#00CC00",
        "#CCCC00",
        "#0000CC",
        "#CC00CC",
        "#00CCCC",
        "#F9EAD6",
        "#3F3F3F",
        "#FF0000",
        "#00FF00",
        "#FFFF00",
        "#0000FF",
        "#FF00FF",
        "#00FFFF",
        "#FFFFFF",
    ),
    "linux": (
        "#000000",
        "#A80000",
        "#00A800",
        "#A85400",
        "#0000A8",
        "#A800A8",
        "#00A8A8",
        "#A8A8A8",
        "#545454",
        "#FF5454",
        "#54FF54",
        "#FFFF54",
        "#5454FF",
        "#FF54FF",
        "#54FFFF",
        "#FFFFFF",
    ),
}


def get_palette(scheme):
    """Get color palette"""
    if scheme in PALETTE_SCHEMES:
        colors = PALETTE_SCHEMES[scheme]

    elif scheme == "custom":
        colors = [eeterm_config.get("terminal", f"palette_{i}") for i in range(0, 16)]
    else:
        return []

    palette = [Gdk.RGBA(*hex2rgba(color)) for color in colors]

    return palette


def get_colors(scheme):
    """Get colors"""
    if scheme in COLOR_SCHEMES:
        foreground, background, color_bold = COLOR_SCHEMES[scheme]

    elif scheme == "custom":
        foreground = eeterm_config.get("terminal", "foreground")
        background = eeterm_config.get("terminal", "background")
        color_bold = eeterm_config.get("terminal", "color_bold")

    else:
        return (None, None, None)

    colors = [
        Gdk.RGBA(*hex2rgba(foreground)),
        Gdk.RGBA(*hex2rgba(background)),
        Gdk.RGBA(*hex2rgba(color_bold)),
    ]

    return colors


def hex2rgba(hex_color, alpha=1.0):
    """Convert hexadecimal color to RGBA"""
    red = int(hex_color[1:3], 16) / 255.0
    green = int(hex_color[3:5], 16) / 255.0
    blue = int(hex_color[5:7], 16) / 255.0
    return (red, green, blue, alpha)


def rgb2hex(red, green, blue):
    """Convert RGB color to hexadecimal"""
    red = int(red * 255)
    green = int(green * 255)
    blue = int(blue * 255)
    return f"#{red:02X}{green:02X}{blue:02X}"


class Terminal(Vte.Terminal):
    """Vte Terminal class for EeTerm"""

    def __init__(self, builder, client=None, data=None, working_directory=None):
        # Calling inherited
        super().__init__()

        self.builder = builder

        self.child_pid = None

        if client is None and data:
            kwargs = {}
            for option in ("agent_forwarding", "compress", "x11_forwarding"):
                value = data.pop(option, None)
                if value is None:
                    value = eessh_config.get("connection", option)

                kwargs[option] = value

            kwargs["no_firsthop"] = data.pop("no_firsthop", False)
            kwargs["username"] = data.pop("username", None)

            client = EeSSHClient(data, embedded=True, **kwargs)

        if client:
            self.label = Gtk.Label(client.title)
        else:
            self.label = Gtk.Label("Terminal")
            if working_directory is None:
                working_directory = os.getcwd()

        self.__client = client
        self.__working_directory = working_directory
        self.__font = None

    @property
    def afwd(self):
        """Check if SSH Agent request forwarding handler is opened"""
        if self.is_local:
            return None

        return self.__client.afwd

    @property
    def broadcast(self):
        """Get broadcast current mode"""
        if self.is_local:
            return "off"

        return self.__client.broadcast

    @property
    def is_local(self):
        """Check if it's a local terminal"""
        return self.__client is None

    @property
    def mode(self):
        """Get current mode (shell/sftp)"""
        if self.is_local:
            return "local"

        return self.__client.mode

    @property
    def pfwds(self):
        """Opened port forwarding tunnels"""
        if self.is_local:
            return None

        return self.__client.pfwds

    @property
    def rfwds(self):
        """Opened reverse port forwarding tunnels"""
        if self.is_local:
            return None

        return self.__client.rfwds

    @property
    def xfwd(self):
        """Get X11 forwarding tunnel display"""
        if self.is_local:
            return None

        return self.__client.xfwd

    def become_root(self):
        """Try to gain root privileges on server"""

        if self.is_local:
            return False

        return self.__client.do_su()

    def close_pfwd(self, *args, callback=None):
        """Close opened port forwarding tunnel with remote"""
        if self.is_local:
            return False

        self.__client.switch_buffer()

        result = self.__client.close_pfwd(*args)
        if callback:
            callback(result)

        self.__client.switch_buffer()

        return result

    def close_rfwd(self, *args, callback=None):
        """Close opened reverse port forwarding tunnel with remote"""
        if self.is_local:
            return False

        self.__client.switch_buffer()

        result = self.__client.close_rfwd(*args)
        if callback:
            callback(result)

        self.__client.switch_buffer()

        return result

    def load_config(self):
        """Load configuration"""
        # General
        if not eeterm_config.get("terminal", "use_system_font"):
            self.__font = Pango.FontDescription(eeterm_config.get("terminal", "font"))
        else:
            self.__font = None

        self.set_font(self.__font)

        self.set_audible_bell(eeterm_config.get("terminal", "audible_bell"))
        self.set_allow_bold(eeterm_config.get("terminal", "allow_bold"))
        self.set_cursor_blink_mode(
            CURSOR_BLINK_MODES[eeterm_config.get("terminal", "cursor_blink_mode")]
        )
        self.set_mouse_autohide(eeterm_config.get("terminal", "mouse_autohide"))
        if hasattr(self, "set_word_char_exceptions"):
            self.set_word_char_exceptions(eeterm_config.get("terminal", "word_char_exceptions"))

        # Colors
        # Palette
        scheme = eeterm_config.get("terminal", "palette_scheme")
        palette = get_palette(scheme)

        # Foreground / Background
        scheme = eeterm_config.get("terminal", "color_scheme")
        foreground, background, color_bold = get_colors(scheme)

        if palette:
            if not foreground:
                foreground = palette[7]
            if not background:
                background = palette[0]

        self.set_colors(foreground, background, palette)
        self.set_color_bold(color_bold)

        # Scrolling
        self.set_scrollback_lines(eeterm_config.get("terminal", "scrollback_lines"))
        self.set_scroll_on_output(eeterm_config.get("terminal", "scroll_on_output"))
        self.set_scroll_on_keystroke(eeterm_config.get("terminal", "scroll_on_keystroke"))

        # Compatibility
        self.set_backspace_binding(
            ERASE_BINDINGS[eeterm_config.get("terminal", "backspace_binding")]
        )
        self.set_delete_binding(ERASE_BINDINGS[eeterm_config.get("terminal", "delete_binding")])

    def copy(self):
        """Return a copy of this terminal"""

        if self.is_local:
            cwd = self.get_current_directory_uri()
            if cwd:
                cwd = GLib.filename_from_uri(cwd)[0]
            return Terminal(self.builder, working_directory=cwd)

        return Terminal(self.builder, client=self.__client.copy())

    def on_activate(self):
        """Callback on terminal starting"""
        self.load_config()

        # Custom scroll
        self.connect("scroll-event", self.on_scroll_event)

        # To set title
        self.connect("window-title-changed", self.on_window_title_changed)

    def on_child_quit(self):
        """Callback on child quit"""
        if not self.is_local:
            return self.__client.close()

        if self.child_pid is None:
            return False

        try:
            os.kill(self.child_pid, signal.SIGHUP)
        except OSError:
            return False

        return True

    def on_scroll_event(self, terminal, event):  # pylint: disable=unused-argument
        """Callback on scroll event"""
        _, delta_x, delta_y = event.get_scroll_deltas()

        state = event.get_state()
        if state & Gdk.ModifierType.CONTROL_MASK and delta_y != 0:
            # Change font size
            if delta_y > 0:
                self.zoom_out()
            else:
                self.zoom_in()

            # Stop propagation so that internal VTE_TERMINAL scrolling is bypassed
            return True

        if state & Gdk.ModifierType.SHIFT_MASK:
            factor = eeterm_config.get("terminal", "scroll_page_increment")
        else:
            factor = eeterm_config.get("terminal", "scroll_step_increment")

        if delta_x:
            # Horizontal adjustment
            adj = self.get_hadjustment()
            adj.set_value(adj.get_value() + delta_x * factor)

        if delta_y:
            # Vertical adjustment
            adj = self.get_vadjustment()
            adj.set_value(adj.get_value() + delta_y * factor)

        # Stop propagation so that internal VTE_TERMINAL scrolling is bypassed
        return True

    def on_window_title_changed(self, terminal):  # pylint: disable=unused-argument
        """Callback on window title change"""
        title = self.get_window_title()
        self.label.set_text(title)

    def open_pfwd(self, *args, callback=None):
        """Wrapper method to client"""

        if self.is_local:
            return False

        self.__client.switch_buffer()

        result = self.__client.open_pfwd(*args)
        if callback:
            callback(result)

        self.__client.switch_buffer()

        return result

    def open_rfwd(self, *args, callback=None):
        """Wrapper method to client"""

        if self.is_local:
            return False

        self.__client.switch_buffer()

        result = self.__client.open_rfwd(*args)
        if callback:
            callback(result)

        self.__client.switch_buffer()

        return result

    def send_files(self, files):
        """Start a ZModem process using zssh to send files"""

        self.__client.restore_tty()

        # Open alternate buffer
        self.__client.stdout.write(b"\x1b[?1049h\x1b[H")

        self.__client.open_sftp()

        def callback(filename, transferred, total, counts):
            print(filename, transferred, total, counts)

        for filepath in files:
            if not self.__client.do_put(filepath, callback=callback):
                pass

        self.__client.close_sftp()

        # Go back to initial buffer
        self.__client.stdout.write(b"\x1b[?1049l")

        self.__client.prepare_tty()

        self.__client.open_shell()

        return True

    def start(self):
        """Start the terminal"""
        if self.is_local:
            # Shell
            argv = eeterm_config.get("terminal", "prefered_shell").split()

            # Get a pty to interact with shell
            pty_ = self.pty_new_sync(Vte.PtyFlags.DEFAULT)
            self.set_pty(pty_)

            def child_setup(pty_):
                pty_.child_setup()

            # Launch shell command
            child_pid, *_ = GLib.spawn_async(
                argv=argv,
                working_directory=self.__working_directory,
                flags=(GLib.SpawnFlags.SEARCH_PATH | GLib.SpawnFlags.DO_NOT_REAP_CHILD),
                child_setup=child_setup,
                user_data=pty_,
            )

            log.debug("vte spawned child process with pid: %s", child_pid)

            self.child_pid = child_pid

            # Will emit «child-exited» signal
            self.watch_child(child_pid)

            GLib.spawn_close_pid(child_pid)

        else:
            # Client
            # Get a pty to interact with client
            master, slave = pty.openpty()

            pty_ = Vte.Pty.new_foreign_sync(master)
            self.set_pty(pty_)

            def on_exit(client, result):  # pylint: disable=unused-argument
                def emit_signal():
                    pty_.close()
                    self.emit("child-exited", 0 if result else 1)

                # Wait a bit before closing pty and sending child-exited signal
                GLib.timeout_add(100, emit_signal)

            def on_ready(client):  # pylint: disable=unused-argument
                def emit_signal():
                    self.emit("child-ready")

                GLib.idle_add(emit_signal)

            # Start client in thread
            thread = threading.Thread(
                target=self.__client.run,
                args=("shell",),
                kwargs={"on_exit": on_exit, "on_ready": on_ready, "tty_fd": slave},
            )
            thread.daemon = True
            thread.start()

        self.on_activate()

        return True

    def toggle_broadcast(self):
        """Toggle broadcast mode"""

        if self.is_local:
            return False

        return self.__client.do_broadcast("toggle")

    def zoom_in(self):
        """Increase zoom level by 10%"""

        font_desc = self.get_font()
        font_size = font_desc.get_size()

        font_size += Pango.SCALE
        font_desc.set_size(font_size)

        self.set_font(font_desc)

    def zoom_out(self):
        """Increase zoom level by 10%"""

        font_desc = self.get_font()
        font_size = font_desc.get_size()

        font_size -= Pango.SCALE
        font_desc.set_size(font_size)

        self.set_font(font_desc)

    def zoom_reset(self):
        """Reset zoom level"""

        self.set_font(self.__font)


#
# Signals
#
GObject.signal_new(
    "child-ready", Terminal, GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, GObject.TYPE_NONE, ()
)
